/* boot-anim-start
 * Copyright 2009 Bobby Powers
 * Copyright (C) 2009 One Laptop per Child
 * Released under the GPLv2+, see COPYING for details
 *
 * starts our graphical boot, but only if the DCON is frozen
 */
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>

#include "utils.h"

static int calc_steps (struct fbinfo *fbinfo);

static void
sighandler (int signo)
{
  /* SIGUSR1 means that the child has finished setting everything up */
  _exit (0);
}

int
main (int argc, char *argv[])
{
  sigset_t mask;
  struct sigaction sa;
  struct fbinfo info;
  int r;

  /* we only animate for prettyboot, which freezes the DCON first */
  if (!dcon_frozen ()) return 0;

  if (fb_open(&info))
    return 1;

  /* this script seems to get called twice (at the beginning of each
   * runlevel), and messes up our nicely choreographed boot animation
   * if we restart the animation the second time it was called. */
  if (boot_anim_started (&info))
    goto err;

  set_started (&info);

  if (calc_steps (&info) < 0)
    goto err;

  fb_close(&info);

  memset(&sa, 0, sizeof (sa));
  sa.sa_handler = sighandler;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGUSR1, &sa, NULL);

  sigemptyset(&mask);
  sigaddset(&mask, SIGUSR1);
  sigprocmask(SIG_BLOCK, &mask, NULL);

  r = fork();
  if (r < 0) {
    return 1;
  } else if (r == 0) {
    /* child */
    /* unblock SIGUSR1 */
    sigprocmask(SIG_UNBLOCK, &mask, NULL);
    execl("/usr/sbin/boot-anim-helper", "boot-anim-helper", NULL);
    _exit (1);
  }

  /* parent */

  /* wait for SIGUSR1 from the child, to indicate that we're ready to
   * continue with boot. use 30-second timeout just in case something goes
   * wrong. */
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
  sleep(30);

  return 1; /* the normal exit path is through _exit(0) inside sighandler */

err:
  fb_close(&info);
  return 1;
}

/* Calculate the number of steps that we should render the animation for.
 * This is based on the 4 initial RC state updates that come in
 * (RCkernelparam, RChostname, RCmountfs and RCswap) plus the number of services
 * that get started in runlevel 5. */
static int
calc_steps (struct fbinfo *fbinfo)
{
  DIR *dir = opendir("/etc/rc.d/rc5.d");
  struct dirent *ent;

  /* initialize at 1, calculated from:
   * 4 initial RC state updates that come in addition to services
   * minus 3 ignored services that come in runlevel 5
   * (0-boot-anim-start, z-boot-anim-stop and ul-warning) */
  int ctr = 4 - 3;

  if (!dir) {
    write (1, "Couldn't open rc5.d directory.\n", 31);
    return -1;
  }

  while ((ent = readdir(dir)))
    if (ent->d_name[0] == 'S')
      ctr++;

  fbinfo->state->step = 0;
  fbinfo->state->total_steps = ctr;
  return 0;
}


