/* boot-anim-client
 * copyright 2009 Bobby Powers, released under the GPLv2+
 * see COPYING for details
 *
 * follows the calling convention of rhgb-client
 */
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "utils.h"

static
uint16_t *draw_frame (struct fbinfo *fbinfo, uint16_t *info)
{
  /* skip the "D565" header */
  info = &info[2];

  while (!is_header (info))
  {
    uint16_t line = info[LINE];
    uint16_t first_col = info[FIRSTCOL];
    uint16_t len = info[LEN];

    if (fbinfo->bpp == 16) {
      memcpy (&fbinfo->fb16[FB_WIDTH*line + first_col], &info[DATA], len*2);
    } else {
      /* re-encode as 32bpp */
      uint16_t *src = &info[DATA];
      uint32_t *dst = &fbinfo->fb32[FB_WIDTH*line + first_col];
      int i;
      for (i = 0; i < len; i++)
        dst[i] = TOARGB(src[i]);
    }

    info = &info[DATA + len];
  }

  return info;
}

int
main (int argc, char *argv[])
{
  struct fbinfo fbinfo;
  uint16_t *info;
  float progress_fraction;
  float frame_position;
  unsigned int frames_to_draw = 1;
  unsigned int show_frame;
  unsigned int next_to_draw;

  /* if we're not called with 'start' as an arg, quit */
  if (argc < 2) return 0;

  /* support plymouth --hide-splash */
  if (strcmp (argv[1], "--hide-splash") == 0) {
    execl ("/usr/sbin/boot-anim-stop", "boot-anim-stop", NULL);
    return 1;
  }

  /* only animate if we're getting an update argument */
  if (strlen (argv[1]) < 8 || strncmp (argv[1], "--update", 8)) return 0;

  /* copied from cscott's original boot-anim-client */
  if (!strcmp (argv[1], "--update=0-boot-anim-start") ||
      !strcmp (argv[1], "--update=z-boot-anim-stop") ||
      !strcmp (argv[1], "--update=ul-warning") ||
      !strcmp (argv[1], "--update=vprocunhide")) return 0;

  /* update animation frame here, if we meet the requirements. */
  if (fb_open(&fbinfo))
    return 1;

  /* make sure that we're animating */
  if (!boot_anim_started (&fbinfo) || !fbinfo.state->running)
    goto cleanup;

  if (fbinfo.state->frame_num >= LAST_FRAME) goto cleanup;

  /* jump to the beginning of the next diff */
  info = &fbinfo.fb16[fbinfo.state->offset];

  /* looking for 'D565' */
  if (!is_header (info)) goto cleanup;

  /* based on the number of times we expected to be invoked (the number of
   * "steps"), we have to decide if we want to move onto the next frame in
   * the animation or not. if there are a low number of total steps, we might
   * even need to render more than one frame.*/

  fbinfo.state->step++;
  if (fbinfo.state->total_steps == 0)
    goto cleanup; /* should never happen */

  /* on a scale of 0 to 1, how far through the animation should we be? */
  progress_fraction = (float) fbinfo.state->step / (float) fbinfo.state->total_steps;

  /* how does that fraction relate to a position in our sequence of frames? */
  frame_position = progress_fraction * (LAST_FRAME + 1);

  /* OK, which frame should we be showing on screen at this step? */
  show_frame = (int) (frame_position + 0.5); /* cheap man's round */
  show_frame--; /* because it's zero-based */

  /* the frame we want to draw might already be on-screen */
  if (show_frame <= fbinfo.state->frame_num)
    goto cleanup;

  if (show_frame > LAST_FRAME)
    show_frame = LAST_FRAME; /* should never happen */

  /* the next frame in the sequence from the one that is on-screen */
  next_to_draw = fbinfo.state->frame_num + 1;

  /* we might have to skip a frame or two, however we can't just skip them
   * because these deltas only render correctly if they are all rendered,
   * in series. so we just end up drawing multiple frames on this invocation. */
  if (show_frame > next_to_draw)
    frames_to_draw += show_frame - next_to_draw;

  while (frames_to_draw-- > 0)
    info = draw_frame(&fbinfo, info);

  fbinfo.state->frame_num = show_frame;
  fbinfo.state->offset = info - fbinfo.fb16;

  if (fbinfo.state->frame_num == LAST_FRAME)
    dcon_freeze (1);

cleanup:
  fb_close(&fbinfo);
  return 0;
}

